import copy
import typing
from typing import Any, Dict, Iterable, List, Optional

from .enums import StandardSqlTypeNames


class StandardSqlDataType:
    def __init__(
            self,
            type_kind: Optional[
                StandardSqlTypeNames
            ] = StandardSqlTypeNames.TYPE_KIND_UNSPECIFIED
    ):
        self._properties: Dict[str, Any] = {}

        self.type_kind = type_kind

    @property
    def type_kind(self) -> Optional[StandardSqlTypeNames]:
        kind = self._properties["typeKind"]
        return StandardSqlTypeNames[kind]

    @type_kind.setter
    def type_kind(self, value: Optional[StandardSqlTypeNames]):
        if not value:
            kind = StandardSqlTypeNames.TYPE_KIND_UNSPECIFIED.value
        else:
            kind = value.value
        self._properties["typeKind"] = kind

    def to_api_repr(self) -> Dict[str, Any]:
        return copy.deepcopy(self._properties)

    @classmethod
    def from_api_repr(cls, resource: Dict[str, Any]):
        type_kind = resource.get("typeKind")
        if type_kind not in StandardSqlTypeNames.__members__:
            type_kind = StandardSqlTypeNames.TYPE_KIND_UNSPECIFIED
        else:
            type_kind = StandardSqlTypeNames[
                typing.cast(str, type_kind)
            ]

        return cls(type_kind)

    def __eq__(self, other):
        if not isinstance(other, StandardSqlDataType):
            return NotImplemented
        else:
            return (
                    self.type_kind == other.type_kind
                    and self.array_element_type == other.array_element_type
                    and self.struct_type == other.struct_type
            )

    def __str__(self):
        result = f"{self.__class__.__name__}(type_kind={self.type_kind!r}, ...)"
        return result


class StandardSqlField:

    def __init__(
            self, name: Optional[str] = None, type: Optional[StandardSqlDataType] = None
    ):
        type_repr = None if type is None else type.to_api_repr()
        self._properties = {"name": name, "type": type_repr}

    @property
    def name(self) -> Optional[str]:
        return typing.cast(Optional[str], self._properties["name"])

    @name.setter
    def name(self, value: Optional[str]):
        self._properties["name"] = value

    @property
    def type(self) -> Optional[StandardSqlDataType]:
        type_info = self._properties["type"]

        if type_info is None:
            return None

        result = StandardSqlDataType()
        result._properties = typing.cast(Dict[str, Any], type_info)

        return result

    @type.setter
    def type(self, value: Optional[StandardSqlDataType]):
        value_repr = None if value is None else value.to_api_repr()
        self._properties["type"] = value_repr

    def to_api_repr(self) -> Dict[str, Any]:
        return copy.deepcopy(self._properties)

    @classmethod
    def from_api_repr(cls, resource: Dict[str, Any]):
        result = cls(
            name=resource.get("name"),
            type=StandardSqlDataType.from_api_repr(resource.get("type", {})),
        )
        return result

    def __eq__(self, other):
        if not isinstance(other, StandardSqlField):
            return NotImplemented
        else:
            return self.name == other.name and self.type == other.type


class StandardSqlTableType:

    def __init__(self, columns: Iterable[StandardSqlField]):
        self._properties = {"columns": [col.to_api_repr() for col in columns]}

    @property
    def columns(self) -> List[StandardSqlField]:
        result = []

        for column_resource in self._properties.get("columns", []):
            column = StandardSqlField()
            column._properties = column_resource  # We do not use a copy on purpose.
            result.append(column)

        return result

    @columns.setter
    def columns(self, value: Iterable[StandardSqlField]):
        self._properties["columns"] = [col.to_api_repr() for col in value]

    def to_api_repr(self) -> Dict[str, Any]:
        return copy.deepcopy(self._properties)

    @classmethod
    def from_api_repr(cls, resource: Dict[str, Any]) -> "StandardSqlTableType":
        columns = []

        for column_resource in resource.get("columns", []):
            type_ = column_resource.get("type")
            if type_ is None:
                type_ = {}

            column = StandardSqlField(
                name=column_resource.get("name"),
                type=StandardSqlDataType.from_api_repr(type_),
            )
            columns.append(column)

        return cls(columns=columns)

    def __eq__(self, other):
        if not isinstance(other, StandardSqlTableType):
            return NotImplemented
        else:
            return self.columns == other.columns



import os.path

import io
import string
import base64
from time import sleep

import oss2
import pyarrow as pa
import logging
import enum
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, FIRST_COMPLETED

from qcloud_cos import CosConfig, CosS3Client

logger = logging.getLogger(__name__)


class QueryDataType(enum.Enum):
    Memory = 0
    File = 1


class ObjectStorageType(enum.Enum):
    OSS = 0
    COS = 1


class QueryData(object):
    def __init__(self, data: list, data_type: QueryDataType, file_list: list = None,
                 object_storage_type: ObjectStorageType = None, oss_bucket: oss2.Bucket = None,
                 cos_client: CosS3Client = None):
        self.data = data
        self.data_type = data_type
        self.file_list = file_list
        self.object_storage_type = object_storage_type
        self.oss_bucket = oss_bucket
        self.cos_client = cos_client
        self.current_file_index = 0
        self.memory_read = False

    def read(self):
        if self.data_type == QueryDataType.Memory:
            if self.memory_read:
                return None
            self.memory_read = True
            return self.data
        elif self.data_type == QueryDataType.File:
            assert self.file_list is not None
            try:
                if self.current_file_index >= len(self.file_list):
                    return None
                file = self.file_list[self.current_file_index]
                file_info = file.split('/', 3)
                result = []
                if self.object_storage_type == ObjectStorageType.OSS:
                    stream = self.oss_bucket.get_object(file_info[3]).read()
                    with pa.ipc.RecordBatchStreamReader(stream) as reader:
                        self.current_file_index += 1
                        for index, row in reader.read_pandas().iterrows():
                            result.append(tuple(row.to_list()))
                        return result
                elif self.object_storage_type == ObjectStorageType.COS:
                    with self.cos_client.get_object(Bucket=file_info[2], Key=file_info[3])[
                        'Body'].get_raw_stream() as stream:
                        with pa.ipc.RecordBatchStreamReader(stream) as reader:
                            self.current_file_index += 1
                            for index, row in reader.read_pandas().iterrows():
                                result.append(tuple(row.to_list()))
                            return result
            except Exception as e:
                logger.error(f'Error while converting from arrow to result: {e}')
                raise Exception(f'Error while converting from arrow to result: {e}')


class Field(object):
    def __init__(self):
        self.name = None
        self.field_type = None
        self.precision = None
        self.scale = None
        self.length = None
        self.nullable = None

    def set_name(self, name):
        self.name = name

    def set_type(self, type):
        self.field_type = type

    def set_precision(self, precision):
        self.precision = precision

    def set_scale(self, scale):
        self.scale = scale

    def set_length(self, length):
        self.length = length

    def set_nullable(self, nullable):
        self.nullable = nullable


class QueryResult(object):
    def __init__(self, total_msg):
        self.data = None
        self.state = None
        self.total_row_count = 0
        self.total_msg = total_msg
        self.schema = []
        self._parse_result_data()

    def _parse_field(self, field: str, schema_field: Field):
        schema_field.set_name(field['name'])
        if field['type'].__contains__('charTypeInfo'):
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) != 'False')
            schema_field.set_length(field['type']['charTypeInfo']['length'])
        elif field['type'].__contains__('decimalTypeInfo'):
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) == 'true')
            schema_field.set_precision(field['type']['decimalTypeInfo']['precision'])
            schema_field.set_scale(field['type']['decimalTypeInfo']['scale'])
        else:
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) == 'true')

    def get_result_state(self) -> string:
        return self.total_msg['status']['state']

    def get_arrow_result(self, arrow_buffer):
        try:
            buffer = base64.b64decode(arrow_buffer)
            with pa.ipc.RecordBatchStreamReader(io.BytesIO(buffer)) as reader:
                table = reader.read_all()
                column_dict = {}
                for index, column_name in enumerate(table.column_names):
                    if column_name in column_dict:
                        column_dict[f'{column_name}_{index}'] = index
                    else:
                        column_dict[column_name] = index
                new_table = table.rename_columns(column_dict.keys())
                pandas_result = new_table.to_pandas()
                result = []
                for index, row in pandas_result.iterrows():
                    result.append(tuple(row.tolist()))
                return result

        except Exception as e:
            logger.error(f'Error while converting from arrow to result: {e}')
            raise Exception(f'Error while converting from arrow to result: {e}')

    def get_result_schema(self):
        fields = self.total_msg['resultSet']['metadata']['fields']
        for field in fields:
            schema_field = Field()
            self._parse_field(field, schema_field)
            self.schema.append(schema_field)

    def get_object_storage_type(self, path: str) -> ObjectStorageType:
        if path.lower().startswith('oss://'):
            return ObjectStorageType.OSS
        elif path.lower().startswith('cos://'):
            return ObjectStorageType.COS

    def get_object_storage_file_list(self) -> list:
        object_storage_file_list = self.total_msg['resultSet']['location']['location']
        return object_storage_file_list

    def get_oss_bucket(self) -> oss2.Bucket:
        location_info = self.total_msg['resultSet']['location']
        id = location_info['stsAkId']
        secret = location_info['stsAkSecret']
        token = location_info['stsToken']
        endpoint = location_info['ossEndpoint']
        if len(location_info['location']) == 0:
            raise Exception('No file found in oss when get result from clickzetta')
        bucket_name = location_info['location'][0].split('/', 3)[2]
        auth = oss2.StsAuth(id, secret, token)
        bucket = oss2.Bucket(auth, endpoint, bucket_name)
        return bucket

    def get_cos_client(self) -> CosS3Client:
        location_info = self.total_msg['resultSet']['location']
        region = location_info['objectStorageRegion']
        id = location_info['stsAkId']
        secret = location_info['stsAkSecret']
        token = location_info['stsToken']
        cos_config = CosConfig(Region=region, SecretId=id, SecretKey=secret, Token=token)
        client = CosS3Client(cos_config)
        return client

    def _parse_result_data(self):
        if len(self.total_msg) == 0:
            return
        self.state = self.total_msg['status']['state']
        if self.state != 'FAILED':
            if 'data' not in self.total_msg['resultSet']:
                if 'location' in self.total_msg['resultSet']:
                    self.get_result_schema()
                    file_list = self.get_object_storage_file_list()
                    if len(file_list) == 0:
                        self.total_row_count = 0
                        self.data = QueryData(data=[], data_type=QueryDataType.Memory)
                        return
                    object_storage_type = self.get_object_storage_type(file_list[0])
                    if object_storage_type == ObjectStorageType.OSS:
                        self.data = QueryData(data_type=QueryDataType.File, file_list=file_list, data=None,
                                              object_storage_type=object_storage_type, oss_bucket=self.get_oss_bucket())
                    elif object_storage_type == ObjectStorageType.COS:
                        self.data = QueryData(data_type=QueryDataType.File, file_list=file_list, data=None,
                                              object_storage_type=object_storage_type, cos_client=self.get_cos_client())

                else:
                    field = Field()
                    field.set_name('RESULT_MESSAGE')
                    field.set_type("STRING")
                    self.schema.append(field)
                    self.total_row_count = 1
                    result_data = [['OPERATION SUCCEED']]
                    self.data = QueryData(data=result_data, data_type=QueryDataType.Memory)
            else:
                if not (len(self.total_msg['resultSet']['data']['data'])):
                    self.total_row_count = 0
                    fields = self.total_msg['resultSet']['metadata']['fields']
                    for field in fields:
                        schema_field = Field()
                    self._parse_field(field, schema_field)
                    self.schema.append(schema_field)
                    self.data = QueryData(data=[], data_type=QueryDataType.Memory)
                    return
                result_data = self.total_msg['resultSet']['data']['data']
                self.get_result_schema()
                query_result = []
                for row in result_data:
                    partial_result = self.get_arrow_result(row)
                    query_result.extend(entity for entity in partial_result)
                self.data = QueryData(data=query_result, data_type=QueryDataType.Memory)

        else:
            raise Exception('SQL job execute failed.Error:' + self.total_msg['status']['message'].split('\n')[0])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: job_result_cache_meta.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bjob_result_cache_meta.proto\x12\x17\x63om.clickzetta.jobcache\"\xa8\x02\n\x12JobResultCacheMeta\x12\x18\n\x0binstance_id\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x19\n\x0cworkspace_id\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x1a\n\rjob_signature\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x1e\n\x11preprocessed_plan\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x19\n\x0cresult_files\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x1a\n\rstart_time_ms\x18\x06 \x01(\x04H\x05\x88\x01\x01\x42\x0e\n\x0c_instance_idB\x0f\n\r_workspace_idB\x10\n\x0e_job_signatureB\x14\n\x12_preprocessed_planB\x0f\n\r_result_filesB\x10\n\x0e_start_time_msB1\n\x1d\x63om.clickzetta.jobcache.protoB\x0eJobCacheProtosP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'job_result_cache_meta_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.clickzetta.jobcache.protoB\016JobCacheProtosP\001'
  _JOBRESULTCACHEMETA._serialized_start=57
  _JOBRESULTCACHEMETA._serialized_end=353
# @@protoc_insertion_point(module_scope)



from clickzetta.client import Client
from clickzetta.enums import LoginParams
from clickzetta.table import Table
from clickzetta.dbapi.connection import connect

__all__ = ["Client", "LoginParams", "Table", "connect"]
